# -*- coding: utf-8 -*-
"""volume discretization structure class
"""
import warnings
import copy

import torch

from torchgdm.constants import DTYPE_FLOAT, DTYPE_COMPLEX, COLORS_DEFAULT
from torchgdm.tools.misc import get_default_device
from torchgdm.struct.base_classes import StructBase
from torchgdm.tools.geometry import get_step_from_geometry
from torchgdm.tools.geometry import sample_random_spherical
from torchgdm.tools.geometry import rotation_x, rotation_y, rotation_z
from torchgdm.tools.misc import ptp
from torchgdm.linearsystem import _reduce_dimensions


# --- base class volume discretized structure container - 3D
class StructDiscretized3D(StructBase):
    """base class volume discretized structure

    Defines the basic volume discretization, its polarizabilities and self-terms
    """

    __name__ = "3D volume discretized structure class"

    def __init__(
        self,
        positions: torch.Tensor,
        materials,
        step=None,
        mesh_normalization_factor: float = 1,
        radiative_correction: bool = False,
        device: torch.device = None,
        **kwargs,
    ):
        """3D discretized structure

        Args:
            positions (torch.Tensor): meshpoint positions (3D, but all y values must be zero)
            materials (material class, or list of them): material of structure. If list, define the material of each meshpoint.
            step (float, optional): nominal stepsize. If not provided, infer automatically from geometry. Defaults to None.
            mesh_normalization_factor (float, optional): mesh normalization. Needs to be adapted for non-cubic meshes. Defaults to 1 (cubic).
            radiative_correction (bool, optional): Add radiative correction term for each meshcell. Defaults to False.
            device (torch.device, optional): Defaults to "cpu" (can be changed by :func:`torchgdm.use_cuda`).

        Raises:
            ValueError: Invalid material config
        """
        super().__init__(
            radiative_correction=radiative_correction, device=device, **kwargs
        )
        self.mesh = "3D"
        self.n_dim = 3

        self.positions = torch.as_tensor(
            positions, dtype=DTYPE_FLOAT, device=self.device
        )
        self.r0 = self.get_center_of_mass()  # center of gravity

        if step is None:
            step_scalar = get_step_from_geometry(self.positions)
        else:
            step_scalar = step
        # step for every meshcell, for consistency with other struct classes
        self.step = step_scalar * torch.ones(
            len(self.positions), dtype=DTYPE_FLOAT, device=self.device
        )

        self.mesh_normalization_factor = torch.as_tensor(
            mesh_normalization_factor, dtype=DTYPE_FLOAT, device=self.device
        )
        if mesh_normalization_factor == 1:
            self.mesh = "cubic"
        else:
            self.mesh = "hexagonal"

        # material of each meshpoint
        if hasattr(materials, "__iter__"):
            if len(materials) != len(self.positions):
                raise ValueError(
                    "Either a global material needs to be given or "
                    + "each meshpoint needs a defined material. "
                    + "But meshpoint list and materials list are of different lengths."
                )
            self.materials = materials
        else:
            self.materials = [materials for i in self.positions]

        self.zeros = torch.zeros(
            (len(self.positions), 3, 3), dtype=DTYPE_COMPLEX, device=self.device
        )

        # discretized, made from natural material: only electric response
        self.evaluation_terms = ["E"]  # possible terms 'E' and 'H'

    def __repr__(self, verbose=False):
        """description about structure"""
        out_str = ""
        out_str += "------ discretized 3D nano-object -------"
        out_str += "\n" + " mesh type:              {}".format(self.mesh)
        out_str += "\n" + " nr. of meshpoints:      {}".format(len(self.positions))
        out_str += "\n" + " nominal stepsizes (nm): {}".format(
            [float(f) for f in torch.unique(self.step)]
        )
        out_str += "\n" + " materials:              {}".format(
            [m.__name__ for m in set(self.materials)]
        )
        bnds = ptp(self.positions, dim=0)
        out_str += "\n" + " size & position:"
        out_str += "\n" + "     X-extension    :    {:.1f} (nm)".format(bnds[0])
        out_str += "\n" + "     Y-extension    :    {:.1f} (nm)".format(bnds[1])
        out_str += "\n" + "     Z-extension    :    {:.1f} (nm)".format(bnds[2])
        out_str += "\n" + "     center of mass :    ({:.1f}, {:.1f}, {:.1f})".format(
            *self.r0
        )

        return out_str

    def set_device(self, device):
        """move all tensors of the class to device"""
        super().set_device(device)

        self.zeros = self.zeros.to(device)
        self.step = self.step.to(device)
        self.r0 = self.r0.to(device)

        self.mesh_normalization_factor = self.mesh_normalization_factor.to(device)

        for mat in self.materials:
            mat.set_device(device)

    # --- self-terms
    def get_selfterm_pE(self, wavelength: float, environment) -> torch.Tensor:
        """return list of 'EE' self-term tensors (3x3) at each meshpoint

        Args:
            wavelength (float): in nm
            environment (environment class): 3D environement class

        Returns:
            torch.Tensor: pE self term tensor
        """
        eps_env = environment.get_environment_permittivity_scalar(
            wavelength, r_probe=self.positions
        )

        norm_nonrad = (
            (-4 * torch.pi)
            * self.mesh_normalization_factor
            / (3 * self.step**3 * eps_env)
        )

        if self.radiative_correction:
            k0 = 2 * torch.pi / wavelength
            norm_rad = (
                self.mesh_normalization_factor
                * (1j * k0**3 * (2 / 3))
                * torch.ones(
                    len(self.positions), dtype=DTYPE_COMPLEX, device=self.device
                )
            )
            cnorm = norm_nonrad + norm_rad
        else:
            cnorm = norm_nonrad

        self_terms_pE = torch.zeros(
            (len(self.positions), 3, 3), dtype=DTYPE_COMPLEX, device=self.device
        )
        self_terms_pE[:, 0, 0] = cnorm
        self_terms_pE[:, 1, 1] = cnorm
        self_terms_pE[:, 2, 2] = cnorm

        return self_terms_pE

    def get_selfterm_mE(self, wavelength: float, environment) -> torch.Tensor:
        return self.zeros

    def get_selfterm_pH(self, wavelength: float, environment) -> torch.Tensor:
        return self.zeros

    def get_selfterm_mH(self, wavelength: float, environment) -> torch.Tensor:
        return self.zeros

    # --- polarizabilities
    def get_polarizability_pE(self, wavelength: float, environment) -> torch.Tensor:
        """return list of EE polarizability tensors (3x3) at each meshpoint

        Args:
            wavelength (float): in nm
            environment (environment class): 3D environement class

        Returns:
            torch.Tensor: pE polarizability tensor
        """
        eps_env = environment.get_environment_permittivity_scalar(
            wavelength, r_probe=self.positions
        )
        eps_env_tensor = torch.zeros(
            (len(self.positions), 3, 3), dtype=DTYPE_COMPLEX, device=self.device
        )
        eps_env_tensor[:, 0, 0] = eps_env
        eps_env_tensor[:, 1, 1] = eps_env
        eps_env_tensor[:, 2, 2] = eps_env

        eps_geo = torch.zeros(
            (len(self.positions), 3, 3), dtype=DTYPE_COMPLEX, device=self.device
        )
        for i, mat in enumerate(self.materials):
            eps_geo[i] = mat.get_epsilon(wavelength)

        vcell_norm = self.step**3 / self.mesh_normalization_factor

        ## --- isotropic polarizability
        alpha_pE = (
            (eps_geo - eps_env_tensor)
            * vcell_norm.unsqueeze(1).unsqueeze(1)
            / (4.0 * torch.pi)
        )

        # with reaction term:
        # k0 = 2 * torch.pi / wavelength
        # alpha_pE = alpha_pE / (1 - (1j * k0**3 * (2 / 3)) * alpha_pE)

        return alpha_pE

    def get_polarizability_mE(self, wavelength: float, environment) -> torch.Tensor:
        return self.zeros

    def get_polarizability_pH(self, wavelength: float, environment) -> torch.Tensor:
        return self.zeros

    def get_polarizability_mH(self, wavelength: float, environment) -> torch.Tensor:
        return self.zeros

    # --- eff. polarizability model extraction wrapper
    def convert_to_effective_polarizability_pair(
        self,
        environment,
        wavelengths: torch.Tensor,
        test_accuracy: bool = False,
        only_pE_mH: bool = True,
        batch_size=16,
        **kwargs,
    ):
        """convert the structure to an effective point polarizability structure model

        The model consists of a pair of electric and magnetic dipole.

        kwargs are passed to :func:`extract_effective_polarizability`

        Args:
            environment (environment class): 3D environement class
            wavelengths (torch.Tensor): list of wavelengths to extract the model at. in nm
            test_accuracy (bool, optional): Whether to test accuracy in a scattering simulation. Defaults to False.
            only_pE_mH (bool, optional): whether to extract only a p/m model (True) or a full (6x6) polarizability (False). Defaults to True.

        Returns:
            :class:`torchgdm.pola.point.StructEffPola3D`: Effective dipole-pair model
        """
        from torchgdm.struct import StructEffPola3D

        wavelengths = torch.as_tensor(
            wavelengths, dtype=DTYPE_FLOAT, device=self.device
        )

        alpha = extract_eff_pola_via_mp_decomposition(
            struct=self,
            environment=environment,
            wavelengths=wavelengths,
            only_pE_mH=only_pE_mH,
            batch_size=batch_size,
            **kwargs,
        )

        struct_aeff = StructEffPola3D(
            positions=torch.stack([self.r0]),
            alpha_dicts=[alpha],
            environment=environment,
            device=self.device,
        )

        # perform a test in an actual scattering simulation
        if test_accuracy:
            test_effective_polarizability_accuracy(struct_aeff, self)

        return struct_aeff

    # --- plotting
    def plot(
        self,
        projection="auto",
        scale=1.0,
        color="auto",
        show_grid=True,
        legend=True,
        set_ax_aspect=True,
        alpha=1.0,
        reset_color_cycle=True,
        **kwargs,
    ):
        """plot the structure (2D)

        Args:
            projection (str, optional): Cartesian projection. Default: "XY" or plane in which all dipoles lie. Defaults to "auto".
            scale (float, optional): scaling factor of the grid cells, if shown. Defaults to 1.0.
            color (str, optional): plot color. Defaults to "auto".
            show_grid (bool, optional): whether to show mesh grid (if available in structure). Defaults to True.
            legend (bool, optional): show legend. Defaults to True.
            set_ax_aspect (bool, optional): automatically set aspect ratio to equal. Defaults to True.
            alpha (int, optional): optional transparency. Defaults to 1.
            reset_color_cycle (bool, optional): reset color cycle after finishing the plot. Defaults to True.

        Returns:
            matplotlib axes
        """
        from torchgdm.visu import visu2d

        im = visu2d.geo2d._plot_structure_discretized(
            self,
            projection=projection,
            scale=scale,
            color=color,
            show_grid=show_grid,
            legend=legend,
            set_ax_aspect=set_ax_aspect,
            alpha=alpha,
            reset_color_cycle=reset_color_cycle,
            **kwargs,
        )
        return im

    def plot_contour(
        self,
        projection="auto",
        color="auto",
        set_ax_aspect=True,
        alpha=1.0,
        alpha_value=None,
        reset_color_cycle=True,
        **kwargs,
    ):
        """plot the contour around the structure (2D)

        Args:
            projection (str, optional): which cartesian plane to project onto. Defaults to "auto".
            color (str, optional): optional matplotlib compatible color. Defaults to "auto".
            set_ax_aspect (bool, optional): If True, will set aspect of plot to equal. Defaults to True.
            alpha (float, optional): matplotlib transparency value. Defaults to 1.0.
            alpha_value (float, optional): alphashape value. If `None`, try to automatically optimize for best enclosing of structure. Defaults to None.
            reset_color_cycle (bool, optional): reset color cycle after finishing the plot. Defaults to True.

        Returns:
            matplotlib line: matplotlib's `scatter` output
        """
        from torchgdm.visu import visu2d

        im = visu2d.geo2d.contour(
            self,
            projection=projection,
            color=color,
            set_ax_aspect=set_ax_aspect,
            alpha=alpha,
            alpha_value=alpha_value,
            reset_color_cycle=reset_color_cycle,
            **kwargs,
        )
        return im

    def plot3d(self, **kwargs):
        """plot the structure (3D)"""
        from torchgdm.visu import visu3d

        return visu3d.geo3d._plot_structure_discretized(self, **kwargs)

    # --- geometry operations
    def rotate(self, alpha, center=torch.as_tensor([0.0, 0.0, 0.0]), axis="z"):
        """rotate the structure

        Args:
            alpha (float): rotation angle (in rad)
            center (torch.Tensor, optional): center of rotation axis. Defaults to torch.as_tensor([0.0, 0.0, 0.0]).
            axis (str, optional): rotation axis, one of ['x', 'y', 'z']. Defaults to 'z'.

        Raises:
            ValueError: unknown rotation axis

        Returns:
            :class:`StructDiscretized3D`: copy of structure with rotated geometry
        """
        _struct_rotated = self.copy()
        center = center.to(dtype=DTYPE_FLOAT, device=self.device)

        if axis.lower() == "x":
            rot = rotation_x(alpha, device=self.device)
        elif axis.lower() == "y":
            rot = rotation_y(alpha, device=self.device)
        elif axis.lower() == "z":
            rot = rotation_z(alpha, device=self.device)
        else:
            raise ValueError("Unknown rotation axis ''.".format(axis))

        if len(_struct_rotated.positions) > 1:
            _struct_rotated.positions = torch.matmul(
                _struct_rotated.positions - (center + self.r0), rot
            ) + (center + self.r0)
        else:
            warnings.warn("Single meshpoint found, ignore rotation.")

        return _struct_rotated


class StructDiscretizedCubic3D(StructDiscretized3D):
    """class for cubic volume discretized structure

    Defines the cubic volume discretization, polarizabilities and self-terms
    """

    __name__ = "3D cubic discretized structure class"

    def __init__(
        self,
        discretization_config,
        step,
        materials,
        device: torch.device = None,
        **kwargs,
    ):
        """3D structure, discretized on a cubic lattice

        Args:
            discretization_config (tuple): tuple of discretization condition function, and discretizer walk limits (as provided by the geometries generators)
            step (float, optional): nominal stepsize. If not provided, infer automatically from geometry. Defaults to None.
            materials (material class, or list of them): material of structure. If list, define the material of each meshpoint.
            device (torch.device, optional): Defaults to "cpu" (can be changed by :func:`torchgdm.use_cuda`).

        Raises:
            ValueError: Invalid material config
        """
        from torchgdm.struct.volume import discretizer_cubic

        positions = discretizer_cubic(*discretization_config, step=step)

        super().__init__(positions, materials, step=step, device=device, **kwargs)

        self.mesh = "cubic"
        self.mesh_normalization_factor = torch.tensor(
            1.0, dtype=DTYPE_FLOAT, device=self.device
        )


class StructDiscretizedHexagonal3D(StructDiscretized3D):
    """class for hexagonal compact volume discretized structure

    Defines the hexagonal compact volume discretization, polarizabilities and self-terms
    """

    __name__ = "3D hexagonal compact discretized structure class"

    def __init__(
        self,
        discretization_config,
        step,
        materials,
        device: torch.device = None,
        **kwargs,
    ):
        """3D structure, discretized on a hexagonal compact lattice

        Args:
            discretization_config (tuple): tuple of discretization condition function, and discretizer walk limits (as provided by the geometries generators)
            step (float, optional): nominal stepsize. If not provided, infer automatically from geometry. Defaults to None.
            materials (material class, or list of them): material of structure. If list, define the material of each meshpoint.
            device (torch.device, optional): Defaults to "cpu" (can be changed by :func:`torchgdm.use_cuda`).

        Raises:
            ValueError: Invalid material config
        """
        from torchgdm.struct.volume import discretizer_hexagonalcompact

        positions = discretizer_hexagonalcompact(*discretization_config, step=step)

        super().__init__(positions, materials, step=step, device=device, **kwargs)

        self.mesh = "hexagonal"
        self.mesh_normalization_factor = torch.sqrt(
            torch.as_tensor(2.0, dtype=DTYPE_FLOAT, device=self.device)
        )


def extract_eff_pola_via_mp_decomposition(
    struct,
    environment,
    wavelengths,
    long_wavelength_approx=False,
    n_dipoles=None,
    distance_dipoles=5000,
    verbose=True,
    only_pE_mH=True,
    progress_bar=True,
    device=None,
    batch_size=16,
    **kwargs,
):
    """Extract effective electric and magnetic dipole polarizability for volume discretized structure

    Extract the polarizability for the structure `struct` in a given `environement`
    at the specified `wavelengths`

    solve inverse problem of adjusting polarizability for different illuminations
    via pseudoinverse

    By default, use 14 plane waves (different incidence directions and polarizations).
    alternative: illumination with `n_dipoles` point-dipole sources if `n_dipoles` is an integer > 0.


    Args:
        struct (class:`StructDiscretized3D`): the discretized structure to extract the model for.
        environment (environment class): 3D environement class.
        wavelengths (torch.Tensor): list of wavelengths to extract the model at. in nm.
        long_wavelength_approx (bool, optional): If True, use long wavelength approximation for dupole extraction. Defaults to False.
        n_dipoles (int, optional): If given, use `n` dipole sources as illumination instead of plane waves. Defaults to None.
        distance_dipoles (int, optional): if using dipoles, specify their distance to the center of gravity. Defaults to 5000.
        verbose (bool, optional): whether to sprint progess info. Defaults to True.
        only_pE_mH (bool, optional): whether to extract only a p/m model (True) or a full (6x6) polarizability (False). Defaults to True.
        progress_bar (bool, optional): whether to show a progress bar. Defaults to True.
        device (str, optional). Defaults to None, in which case the structure's device is used.

    Returns:
        dict: contains all informations about the effective dipole polarizability model
    """
    from torchgdm.env.freespace_3d.inc_fields import ElectricDipole, PlaneWave
    from torchgdm.simulation import Simulation
    from torchgdm.postproc.multipole import decomposition_exact
    from torchgdm.tools.geometry import get_enclosing_sphere_radius
    from torchgdm.tools.misc import tqdm

    if verbose:
        import time
    _struct = struct.copy()

    if device is None:
        device = _struct.device
    else:
        struct.set_device(device)

    # convert single wavelength to list
    wavelengths = torch.as_tensor(wavelengths, dtype=DTYPE_FLOAT, device=device)
    wavelengths = torch.atleast_1d(wavelengths)

    # use first order multipole moments (propto local field)
    which_moment_p = "ed_1"
    which_moment_m = "md"
    enclosing_radius = get_enclosing_sphere_radius(_struct.positions)
    r_sphere = enclosing_radius + distance_dipoles
    r0 = struct.get_center_of_mass()

    # setup perpendicular plane waves illuminations
    if n_dipoles is None:
        pw_conf_list = [
            [0.0, 1.0, 0, "xz"],  # E-x, H-y, k-z
            [1.0, 0.0, 0, "xz"],  # E-y, H-x, k-z
            #
            [1.0, 0.0, torch.pi / 2.0, "xz"],  # E-x, H-z, k-y
            [0.0, 1.0, torch.pi / 2.0, "xz"],  # E-z, H-x, k-y
            #
            [1.0, 0.0, torch.pi / 2.0, "yz"],  # E-y, H-z, k-x
            [0.0, 1.0, torch.pi / 2.0, "yz"],  # E-z, H-y, k-x
            #
            [1.0, 0.0, -torch.pi / 2.0, "xz"],  # E-x, H-z, -k-y
            [0.0, 1.0, -torch.pi / 2.0, "xz"],  # E-z, H-x, -k-y
            #
            [1.0, 0.0, -torch.pi / 2.0, "yz"],  # E-y, H-z, -k-x
            [0.0, 1.0, -torch.pi / 2.0, "yz"],  # E-z, H-y, -k-x
            #
            [1.0, 0.0, torch.pi / 4.0, "xz"],  # oblique
            [0.0, 1.0, torch.pi / 4.0, "yz"],  # oblique
            #
            [0.0, 1.0, -torch.pi / 4.0, "xz"],  # oblique, opposite
            [1.0, 0.0, -torch.pi / 4.0, "yz"],  # oblique, opposite
        ]
        e_inc_list = [
            PlaneWave(e0s=a, e0p=b, inc_angle=c, inc_plane=d, device=device)
            for [a, b, c, d] in pw_conf_list
        ]
    # optional: multiple dipole illuminations
    else:
        if n_dipoles <= 0 or type(n_dipoles) != int:
            raise ValueError(
                "dipole illumination mode: `n_dipoles` needs to be a positive integer."
            )

        # setup dipoles of random position and random orientation
        rnd_pos = sample_random_spherical(n_dipoles) * r_sphere
        e_inc_list = [
            ElectricDipole(
                r_source=r_dp,
                p_source=torch.rand(3, device=device) * r_sphere,
                device=device,
            )
            for r_dp in rnd_pos
        ]

    # replace illumination
    _sim = Simulation(
        structures=[_struct],
        environment=environment,
        illumination_fields=e_inc_list,
        wavelengths=wavelengths,
        device=device,
    )

    if verbose:
        t0 = time.time()
        _pos_p, _pos_m = _sim._get_polarizable_positions_separate_p_m()
        n_dp = len(_pos_p) + len(_pos_m)
        n_wl = len(wavelengths)
        print("Running simulation ({} dipoles, {} wls)... ".format(n_dp, n_wl), end="")
    _sim.run(verbose=False, progress_bar=progress_bar, batch_size=batch_size)

    if verbose and not progress_bar:
        print("Done in {:.2f}s.".format(time.time() - t0))

    # solve the optimization problem
    alpha_6x6 = torch.zeros(
        (len(wavelengths), 6, 6), dtype=DTYPE_COMPLEX, device=device
    )
    if verbose:
        t0 = time.time()
        print("Running p/m optimization... ", end="")

    for i_wl in tqdm(range(len(wavelengths)), progress_bar, title=""):
        wl = wavelengths[i_wl]

        # multipole expansion for all illuminations
        mp_dict = decomposition_exact(
            _sim,
            wl,
            long_wavelength_approx=long_wavelength_approx,
        )
        p_eval = mp_dict[which_moment_p]
        m_eval = mp_dict[which_moment_m]

        # illuminating fields at expansion location
        e0_eval = torch.zeros((len(e_inc_list), 3), dtype=DTYPE_COMPLEX, device=device)
        h0_eval = torch.zeros((len(e_inc_list), 3), dtype=DTYPE_COMPLEX, device=device)
        for i_field, e_inc in enumerate(e_inc_list):
            e0 = e_inc.get_efield(r0.unsqueeze(0), wl, environment)
            e0_eval[i_field] = e0
            h0 = e_inc.get_hfield(r0.unsqueeze(0), wl, environment)
            eps_env = environment.get_environment_permittivity_scalar(wl, r0)
            h0_eval[i_field] = h0

        # --- full 6x6 polarizability
        if not only_pE_mH:
            # pseudo-inverse of all illuminations
            f0_eval = torch.cat([e0_eval, h0_eval], dim=1)
            pinv_f0 = torch.linalg.pinv(f0_eval)

            # optimum alphas to obtain dipole moments for each illumination
            pm_eval = torch.cat([p_eval, m_eval], dim=1)
            alpha_6x6_inv = torch.matmul(pinv_f0, pm_eval)
            alpha_6x6[i_wl] = alpha_6x6_inv

            # test residuals
            _test_residual_effective_polarizability(
                _struct,
                wl,
                environment,
                alpha_6x6_inv,
                f0_eval,
                pm_eval,
                text_which_dp="6x6 dipole",
            )

        # --- only pE and mH
        if only_pE_mH:
            # pseudo-inverse of all illuminations
            pinv_e0 = torch.linalg.pinv(e0_eval)
            pinv_h0 = torch.linalg.pinv(h0_eval)

            # optimum alphas to obtain dipole moments for each illumination
            alpha_pinv = torch.matmul(pinv_e0, p_eval)
            alpha_minv = torch.matmul(pinv_h0, m_eval)

            alpha_6x6[i_wl, :3, :3] = alpha_pinv
            alpha_6x6[i_wl, 3:, 3:] = alpha_minv

            # test residuals
            _test_residual_effective_polarizability(
                _struct,
                wl,
                environment,
                alpha_pinv,
                e0_eval,
                p_eval,
                text_which_dp="electric dipole",
            )
            _test_residual_effective_polarizability(
                _struct,
                wl,
                environment,
                alpha_minv,
                h0_eval,
                m_eval,
                text_which_dp="magnetic dipole",
            )

    if verbose:
        print("Done in {:.2f}s.".format(time.time() - t0))

    dict_pola_pseudo = dict(
        r0=r0,
        r0_MD=r0,
        r0_ED=r0,
        full_geometry=_struct.positions,
        alpha_6x6=alpha_6x6,
        wavelengths=wavelengths,
        enclosing_radius=enclosing_radius,
        k0_spectrum=2 * torch.pi / wavelengths,
    )
    return dict_pola_pseudo


def extract_eff_pola_via_propagation(
    struct,
    environment,
    wavelengths,
    n_probe=150,
    distance_probe=10.0,
    n_dipoles=None,
    distance_dipoles=5000,
    verbose=True,
    only_pE_mH=False,
    progress_bar=True,
    device=None,
):
    """Extract effective electric and magnetic dipole polarizability for volume discretized structure

    Extract the polarizability for the structure `struct` in a given `environement`
    at the specified `wavelengths`

    In this version, the effective dipole response for several illuminations is obtained
    via matching on a circumscribing sphere by solving of a first inverse problem via pseudoinverse.

    The second inverse problem of adjusting polarizability for different
    illuminations is also solved via pseudoinverse.

    By default, use 14 plane waves (different incidence directions and polarizations).
    alternative: illumination with `n_dipoles` point-dipole sources if `n_dipoles` is an integer > 0.


    Args:
        struct (class:`StructDiscretized3D`): the discretized structure to extract the model for.
        environment (environment class): 3D environement class.
        wavelengths (torch.Tensor): list of wavelengths to extract the model at. in nm.
        n_probe (int, optional): number of probe positions on enclosing sphere. Defaults to 100.
        distance_probe (float, optional): additional distance to enclosing sphere in units of discretization step. Defaults to 2.0.
        n_dipoles (int, optional): If given, use `n` dipole sources as illumination instead of plane waves. Defaults to None.
        distance_dipoles (int, optional): if using dipoles, specify their distance to the center of gravity. Defaults to 5000.
        verbose (bool, optional): whether to sprint progess info. Defaults to True.
        only_pE_mH (bool, optional): whether to extract only a p/m model (True) or a full (6x6) polarizability (False). Defaults to True.
        progress_bar (bool, optional): whether to show a progress bar. Defaults to True.
        device (str, optional). Defaults to None, in which case the structure's device is used.

    Returns:
        dict: contains all informations about the effective dipole polarizability model
    """
    from torchgdm.env.freespace_3d.inc_fields import ElectricDipole, PlaneWave
    from torchgdm.simulation import Simulation
    from torchgdm.postproc.fields import nf
    from torchgdm.tools.geometry import get_enclosing_sphere_radius
    from torchgdm.tools.misc import tqdm

    if verbose:
        import time
    _struct = struct.copy()

    if device is None:
        device = _struct.device
    else:
        struct.set_device(device)

    # convert single wavelength to list
    wavelengths = torch.as_tensor(wavelengths, dtype=DTYPE_FLOAT, device=device)
    wavelengths = torch.atleast_1d(wavelengths)

    # use first order multipole moments (propto local field)
    step_max = torch.max(struct.step)
    enclosing_radius = get_enclosing_sphere_radius(_struct.positions)
    illumination_radius = enclosing_radius + distance_dipoles
    probe_radius = enclosing_radius + distance_probe * step_max
    r0 = struct.get_center_of_mass()

    # setup perpendicular plane waves illuminations
    if n_dipoles is None:
        pw_conf_list = [
            [0.0, 1.0, 0, "xz"],  # E-x, H-y, k-z
            [1.0, 0.0, 0, "xz"],  # E-y, H-x, k-z
            #
            [1.0, 0.0, torch.pi / 2.0, "xz"],  # E-x, H-z, k-y
            [0.0, 1.0, torch.pi / 2.0, "xz"],  # E-z, H-x, k-y
            #
            [1.0, 0.0, torch.pi / 2.0, "yz"],  # E-y, H-z, k-x
            [0.0, 1.0, torch.pi / 2.0, "yz"],  # E-z, H-y, k-x
            #
            [1.0, 0.0, -torch.pi / 2.0, "xz"],  # E-x, H-z, -k-y
            [0.0, 1.0, -torch.pi / 2.0, "xz"],  # E-z, H-x, -k-y
            #
            [1.0, 0.0, -torch.pi / 2.0, "yz"],  # E-y, H-z, -k-x
            [0.0, 1.0, -torch.pi / 2.0, "yz"],  # E-z, H-y, -k-x
            #
            [1.0, 0.0, torch.pi / 4.0, "xz"],  # oblique
            [0.0, 1.0, torch.pi / 4.0, "yz"],  # oblique
            #
            [0.0, 1.0, -torch.pi / 4.0, "xz"],  # oblique, opposite
            [1.0, 0.0, -torch.pi / 4.0, "yz"],  # oblique, opposite
        ]
        e_inc_extract = [
            PlaneWave(e0s=a, e0p=b, inc_angle=c, inc_plane=d, device=device)
            for [a, b, c, d] in pw_conf_list
        ]
    # optional: multiple dipole illuminations
    else:
        if n_dipoles <= 0 or type(n_dipoles) != int:
            raise ValueError(
                "dipole illumination mode: `n_dipoles` needs to be a positive integer."
            )

        # setup dipoles of random position and random orientation
        rnd_pos_dp = sample_random_spherical(n_dipoles) * illumination_radius
        e_inc_extract = [
            ElectricDipole(
                r_source=r_dp,
                p_source=torch.rand(3, device=device) * illumination_radius,
                device=device,
            )
            for r_dp in rnd_pos_dp
        ]

    # setup field probe positions on enclosing sphere
    r_probe = sample_random_spherical(n_probe) * probe_radius

    # replace illumination
    _sim = Simulation(
        structures=[_struct],
        environment=environment,
        illumination_fields=e_inc_extract,
        wavelengths=wavelengths,
        device=device,
    )

    if verbose:
        t0 = time.time()
        _pos_p, _pos_m = _sim._get_polarizable_positions_separate_p_m()
        n_dp = len(_pos_p) + len(_pos_m)
        n_wl = len(wavelengths)
        print("Running simulation ({} dipoles, {} wls)... ".format(n_dp, n_wl), end="")
    _sim.run(verbose=False, progress_bar=progress_bar)

    if verbose and not progress_bar:
        print("Done in {:.2f}s.".format(time.time() - t0))

    # solve the optimization problem
    alpha_6x6 = torch.zeros(
        (len(wavelengths), 6, 6), dtype=DTYPE_COMPLEX, device=device
    )
    if verbose:
        t0 = time.time()
        print("Running p/m optimization... ", end="")

    for i_wl in tqdm(range(len(wavelengths)), progress_bar, title=""):
        wl = wavelengths[i_wl]

        # calculate fields on enclosing sphere for all illuminations
        # shape: (n_illumination, n_probe*6)
        nf_probe = nf(_sim, wl, r_probe=r_probe, progress_bar=False)
        e_eval = nf_probe["sca"].get_efield()
        h_eval = nf_probe["sca"].get_hfield()
        f_eval = torch.cat([e_eval, h_eval], dim=2)
        f_eval = f_eval.reshape(len(e_inc_extract), -1)

        # illuminating fields at expansion location
        # shape: (n_illumination, 6)
        e0_eval = torch.zeros(
            (len(e_inc_extract), 3), dtype=DTYPE_COMPLEX, device=device
        )
        h0_eval = torch.zeros(
            (len(e_inc_extract), 3), dtype=DTYPE_COMPLEX, device=device
        )
        for i_field, e_inc in enumerate(e_inc_extract):
            e0 = e_inc.get_efield(r0.unsqueeze(0), wl, environment)
            e0_eval[i_field] = e0
            h0 = e_inc.get_hfield(r0.unsqueeze(0), wl, environment)
            eps_env = environment.get_environment_permittivity_scalar(wl, r0)
            h0_eval[i_field] = h0
        f0_eval = torch.cat([e0_eval, h0_eval], dim=1)

        # --- full 6x6 polarizability
        if not only_pE_mH:
            # calculate Green's tensors between r0 and r_probe
            # shape: (n_probe*6, 6)
            G_6x6 = environment.get_G_6x6(r_probe=r_probe, r_source=r0, wavelength=wl)
            G_all = G_6x6.reshape(-1, 6)

            # inv. problem #1: probe fields + Green's tensors --> dipole moments
            Gi = torch.linalg.pinv(G_all)
            pm_eff = torch.matmul(Gi.unsqueeze(0), f_eval.unsqueeze(-1))[..., 0]

            # inv. problem #2: dipole moments + illumination --> effective pola
            pinv_f0 = torch.linalg.pinv(f0_eval)
            alpha_6x6_inv = torch.matmul(pinv_f0, pm_eff)

            # optimum alphas to obtain dipole moments for each illumination
            alpha_6x6[i_wl] = alpha_6x6_inv

            # test residuals
            _test_residual_effective_polarizability(
                _struct,
                wl,
                environment,
                alpha_6x6_inv,
                f0_eval,
                pm_eff,
                text_which_dp="6x6 dipole",
            )

        # --- only pE and mH
        if only_pE_mH:
            # calculate Green's tensors between r0 and r_probe
            # shape: (n_probe*6, 6)
            G_Ep = environment.get_G_Ep(r_probe=r_probe, r_source=r0, wavelength=wl)
            G_Hm = environment.get_G_Hm(r_probe=r_probe, r_source=r0, wavelength=wl)
            G_Ep_all = G_Ep.reshape(-1, 3)
            G_Hm_all = G_Hm.reshape(-1, 3)

            e_eval = e_eval.reshape(len(e_inc_extract), -1)
            h_eval = h_eval.reshape(len(e_inc_extract), -1)

            # inv. problem #1: probe fields + Green's tensors --> dipole moments
            G_Ep_i = torch.linalg.pinv(G_Ep_all)
            G_Hm_i = torch.linalg.pinv(G_Hm_all)
            p_eff = torch.matmul(G_Ep_i.unsqueeze(0), e_eval.unsqueeze(-1))[..., 0]
            m_eff = torch.matmul(G_Hm_i.unsqueeze(0), h_eval.unsqueeze(-1))[..., 0]

            # pseudo-inverse of all illuminations
            pinv_e0 = torch.linalg.pinv(e0_eval)
            pinv_h0 = torch.linalg.pinv(h0_eval)

            # optimum alphas to obtain dipole moments for each illumination
            alpha_pinv = torch.matmul(pinv_e0, p_eff)
            alpha_minv = torch.matmul(pinv_h0, m_eff)

            alpha_6x6[i_wl, :3, :3] = alpha_pinv
            alpha_6x6[i_wl, 3:, 3:] = alpha_minv

            # test residuals
            _test_residual_effective_polarizability(
                _struct,
                wl,
                environment,
                alpha_pinv,
                e0_eval,
                p_eff,
                text_which_dp="electric dipole",
            )
            _test_residual_effective_polarizability(
                _struct,
                wl,
                environment,
                alpha_minv,
                h0_eval,
                m_eff,
                text_which_dp="magnetic dipole",
            )

    if verbose:
        print("Done in {:.2f}s.".format(time.time() - t0))

    dict_pola_pseudo = dict(
        r0=r0,
        r0_MD=r0,
        r0_ED=r0,
        full_geometry=_struct.positions,
        alpha_6x6=alpha_6x6,
        wavelengths=wavelengths,
        enclosing_radius=enclosing_radius,
        k0_spectrum=2 * torch.pi / wavelengths,
    )

    return dict_pola_pseudo


def _test_residual_effective_polarizability(
    _struct,
    wavelength,
    environment,
    alpha_eff,
    f0,
    dp_moments,
    text_which_dp="",
    residual_warning_threshold=0.25,
):
    # --- test - calculate mean residuals for accuracy estimation
    epsilon_dpm = torch.abs(
        _struct.get_polarizability_6x6(wavelength=wavelength, environment=environment)
    ).max()  # add a single mesh-cell polarizability as epsilon to relative error test

    res_p = torch.abs(dp_moments - torch.matmul(alpha_eff, f0.T).T)
    norm_p = torch.linalg.norm(dp_moments, dim=-1).unsqueeze(1) + epsilon_dpm

    if torch.max(res_p / norm_p) > residual_warning_threshold:
        print(
            "Warning: wl={}nm - eff. {} pola. peak residual is exceeing the threshold! ({:.4f} > {:.2f})".format(
                wavelength,
                text_which_dp,
                torch.max(res_p / norm_p),
                residual_warning_threshold,
            )
        )


def test_effective_polarizability_accuracy(
    struct_alpha,
    struct_full,
    test_yz_incidence=True,
    environment=None,
    rtol=0.10,
    verbose=True,
    progress_bar=False,
    device=None,
):
    """test effective polarizability model in a scattering simulation

    Print some information about model accuracy

    Args:
        struct_alpha (torchgdm.struct.point.StructEffPola3D): effective polarizability structure
        struct_full (StructDiscretized3D): associated full discretization structure as reference
        test_yz_incidence (bool, optional): Whether to test also YZ incident plane. Defaults to True.
        environment (3D env. class, optional): Simulation environment. If None, use environment from effective dipole model structure. Defaults to None.
        rtol (float, optional): relative error threshold for raising warnings. Defaults to 0.10.
        verbose (bool, optional): Print detailed info. Defaults to True.
        progress_bar (bool, optional): Show progress bars. Defaults to False.
        device (str, optional): If None, use structure's device. Defaults to None.

    """
    from torchgdm.tools.batch import calc_spectrum
    from torchgdm.postproc import crosssect
    from torchgdm.env.freespace_3d.inc_fields import PlaneWave
    from torchgdm.simulation import Simulation
    from torchgdm.tools.misc import to_np

    if environment is None:
        if struct_alpha.environment is None:
            raise ValueError(
                "Structure does not contain environement definition, "
                + "and no environemnt has been specified. Please provide the environment."
            )
        environment = struct_alpha.environment

    if device is None:
        device = struct_alpha.device

    wavelengths = struct_alpha.wavelengths_data

    # test configs: plane wave, s/p-polarization, 0/90 deg incidence
    e_inc_list = [
        PlaneWave(e0p=1.0, e0s=0.0, inc_angle=0),
        PlaneWave(e0p=0.0, e0s=1.0, inc_angle=0),
        PlaneWave(e0p=1.0, e0s=0.0, inc_angle=torch.pi / 2),
        PlaneWave(e0p=0.0, e0s=1.0, inc_angle=torch.pi / 2),
    ]
    if test_yz_incidence:
        e_inc_list += [
            PlaneWave(e0p=1.0, e0s=0.0, inc_plane="yz", inc_angle=torch.pi / 2),
            PlaneWave(e0p=0.0, e0s=1.0, inc_plane="yz", inc_angle=torch.pi / 2),
        ]

    # setup a discretized and a effective dipole pair simulation
    sim_alpha = Simulation(
        structures=[struct_alpha],
        environment=environment,
        illumination_fields=e_inc_list,
        wavelengths=wavelengths,
        device=device,
    )
    sim_discr = Simulation(
        structures=[struct_full],
        environment=environment,
        illumination_fields=e_inc_list,
        wavelengths=wavelengths,
        device=device,
    )

    # run simulations and calc. cross section spectra
    if verbose:
        print("-" * 60)
        print("Testing effective polarizability vs. discretized simulation.")

    sim_alpha.run(verbose=False, progress_bar=progress_bar)
    cs_alpha = calc_spectrum(sim_alpha, crosssect.total, progress_bar=progress_bar)

    sim_discr.run(verbose=False, progress_bar=progress_bar)
    cs_discr = calc_spectrum(sim_discr, crosssect.total, progress_bar=progress_bar)

    # calculate errors
    for k in ["scs", "ecs"]:
        try:
            rel_diff = (cs_discr[k] - cs_alpha[k]) / ((cs_discr[k] + cs_alpha[k]))
            mean_rel_error = torch.mean(torch.abs(rel_diff))
            peak_rel_error = torch.max(torch.abs(rel_diff))
            if verbose:
                print("'{}':".format(k))
                print("    - mean rel. error: {:.3f}".format(to_np(mean_rel_error)))
                print("    - peak rel. error: {:.3f}".format(to_np(peak_rel_error)))
            if mean_rel_error > rtol:
                warnings.warn(
                    "'{}': Effective polarizability mean relative error larger than tolerance ({:.3f} > {:.3f})".format(
                        k, mean_rel_error, rtol
                    )
                )

            elif peak_rel_error > rtol:
                warnings.warn(
                    "'{}': Effective polarizability peak relative error larger than tolerance ({:.3f} > {:.3f})".format(
                        k, peak_rel_error, rtol
                    )
                )

        except TypeError:
            pass
    if verbose:
        print("-" * 60)


# %% --- test
if __name__ == "__main__":
    import torchgdm as tg

    device = "cpu"
    # device = "cuda"

    # --- setup test case simulation
    # - illumination field(s)
    wavelengths = torch.tensor([650.0])
    wavelengths = torch.linspace(500, 800, 21)

    # - environment
    eps_env = 1.0
    mat_env = tg.materials.MatConstant(eps_env)
    env = tg.env.freespace_3d.EnvHomogeneous3D(env_material=mat_env)

    # - dummy illumination field
    e_inc_dummy = tg.env.freespace_3d.PlaneWave(inc_angle=torch.pi / 4)

    # - first structure: volume discretization
    l = 8
    w = 4
    h = 4
    step = 20.0
    mat_struct = tg.materials.MatConstant(eps=10.0)
    mat_struct = tg.materials.MatDatabase("Ge")
    geometry = tg.struct.volume.discretizer_cubic(
        *tg.struct.volume.cuboid(l, w, h), step=step
    )
    # geometry = tg.struct.volume.discretizer_hexagonalcompact(
    #     *tg.struct.volume.split_ring(r_in=9, r_out=12, h=2, alpha_g=torch.pi/8), step=step
    # )

    struct_mesh = tg.struct.StructDiscretized3D(geometry, mat_struct)
    # struct_mesh2 = struct_mesh.rotate(torch.pi / 2)
    # struct_mesh = struct_mesh + (struct_mesh2 + [(w) / 2 * step, (l + w) / 2 * step, 0])
    struct_mesh.plot()

    struct_alpha_pE_mH = struct_mesh.convert_to_effective_polarizability_pair(
        env, wavelengths, only_pE_mH=True, test_accuracy=True
    )
    struct_alpha_6x6 = struct_mesh.convert_to_effective_polarizability_pair(
        env, wavelengths, only_pE_mH=False, test_accuracy=True
    )

    # - full simulation
    sim_full = tg.simulation.Simulation(
        structures=[struct_mesh],
        environment=env,
        illumination_fields=[e_inc_dummy],
        wavelengths=wavelengths,
        device=device,
    )

    alpha_pm_test = struct_alpha_pE_mH.get_polarizability_6x6(wavelengths[0], env)
    alpha_66_test = struct_alpha_6x6.get_polarizability_6x6(wavelengths[0], env)
    # %%
    import matplotlib.pyplot as plt

    plt.figure(figsize=(8, 8))
    plt.subplot(221)
    plt.imshow(alpha_pm_test[0].real)
    plt.colorbar()
    plt.subplot(222)
    plt.imshow(alpha_pm_test[0].imag)
    plt.colorbar()

    plt.subplot(223)
    plt.imshow(alpha_66_test[0].real)
    plt.colorbar()
    plt.subplot(224)
    plt.imshow(alpha_66_test[0].imag)
    plt.colorbar()
    plt.show()
    # %%
    sim_pEmH = tg.simulation.Simulation(
        structures=[struct_alpha_pE_mH],
        environment=env,
        illumination_fields=[e_inc_dummy],
        wavelengths=wavelengths,
        device=device,
    )
    sim_6x6 = tg.simulation.Simulation(
        structures=[struct_alpha_6x6],
        environment=env,
        illumination_fields=[e_inc_dummy],
        wavelengths=wavelengths,
        device=device,
    )
    sim_full.run()
    sim_pEmH.run()
    sim_6x6.run()

    cs_f = sim_full.get_spectra_scs()
    cs_pm = sim_pEmH.get_spectra_scs()
    cs_66 = sim_6x6.get_spectra_scs()
    plt.plot(wavelengths, cs_f["scs"], label="scs-full")
    plt.plot(wavelengths, cs_pm["scs"], label="scs-pm")
    plt.plot(wavelengths, cs_66["scs"], label="scs-6x6")
    plt.legend()
    plt.show()
# %%
# print(sim_full.get_spectra_scs())
# print(sim_pEmH.get_spectra_scs())
# print(sim_6x6.get_spectra_scs())

# %%



# encoding=utf-8
"""contains classes necessary for describing simulations in a specific environment:
  - Green's tensors, describing environments and their boundary conditions.
  - illumination fields specific for the environment
"""
from . import freespace_2d
from . import freespace_3d

from .freespace_2d import EnvHomogeneous2D
from .freespace_3d import EnvHomogeneous3D


# -*- coding: utf-8 -*-
"""
base classes for defining torchgdm simulations
"""
# %%
import warnings

import torch

from torchgdm.constants import DTYPE_FLOAT, DTYPE_COMPLEX
from torchgdm.tools.misc import get_default_device


# --- environment defining base class
class EnvironmentBase:
    """base class defining the environment via Green's tensors"""

    __name__ = "environment base class"

    def __init__(self, device: torch.device = None):
        if device is None:
            self.device = get_default_device()
        else:
            self.device = device

        self.n_dim = -1  # problem dimension needs to be set by child class

    def __repr__(self, verbose: bool = False):
        """description about simulation environment defined by set of dyads"""
        out_str = " ------ base environment class - doesn't define anything yet -------"
        return out_str

    def set_device(self, device):
        """move all tensors of the class to device"""
        self.device = device

    def copy(self):
        import copy

        return copy.deepcopy(self)

    def get_environment_permittivity_scalar(self, wavelength: float, r_probe: tuple):
        """return complex environment permittivity tensor - scalar(!)

        evaluate scalar epsilon for `wavelength` at position `r` (both in nm)
        """
        raise NotImplementedError(
            "Definition missing! This function needs to be overridden in child class!"
        )

    def get_G_6x6(
        self, r_probe: torch.Tensor, r_source: torch.Tensor, wavelength: float
    ):
        """electro-magnetic Green's tensor for an e/m pair-dipole"""
        raise NotImplementedError(
            "Definition missing! This function needs to be overridden in child class!"
        )

    def get_G_Ep(
        self, r_probe: torch.Tensor, r_source: torch.Tensor, wavelength: float
    ):
        """Electric field Green's tensor for an electric dipole"""
        raise NotImplementedError(
            "Definition missing! This function needs to be overridden in child class!"
        )

    def get_G_Hp(
        self, r_probe: torch.Tensor, r_source: torch.Tensor, wavelength: float
    ):
        """Magnetic field Green's tensor for an electric dipole"""
        raise NotImplementedError(
            "Definition missing! This function needs to be overridden in child class!"
        )

    def get_G_Em(
        self, r_probe: torch.Tensor, r_source: torch.Tensor, wavelength: float
    ):
        """Electric field Green's tensor for a magnetic dipole"""
        raise NotImplementedError(
            "Definition missing! This function needs to be overridden in child class!"
        )

    def get_G_Hm(
        self, r_probe: torch.Tensor, r_source: torch.Tensor, wavelength: float
    ):
        """Magnetic field Green's tensor for a magnetic dipole"""
        raise NotImplementedError(
            "Definition missing! This function needs to be overridden in child class!"
        )

    # --- wrapper for mixed e-m Green's tensors
    def get_G_6x6(self, r_probe, r_source, wavelength):
        """electro-magnetic Green's tensor for a pair of p/m dipoles

        r_probe: field evaluation position, r_source: emitter-pair position
        """
        G_6x6 = torch.cat(
            [
                torch.cat(
                    [
                        self.get_G_Ep(r_probe, r_source, wavelength),
                        self.get_G_Em(r_probe, r_source, wavelength),
                    ],
                    dim=-1,
                ),
                torch.cat(
                    [
                        self.get_G_Hp(r_probe, r_source, wavelength),
                        self.get_G_Hm(r_probe, r_source, wavelength),
                    ],
                    dim=-1,
                ),
            ],
            dim=-2,
        )
        return G_6x6

    def get_G_EHp_6x3(self, r_probe, r_source, wavelength):
        """electric field Green's tensor for a pair of p/m dipoles

        r_probe: field evaluation position, r_source: emitter-pair position
        """
        G_EHp_6x3 = torch.cat(
            [
                self.get_G_Ep(r_probe, r_source, wavelength),
                self.get_G_Hp(r_probe, r_source, wavelength),
            ],
            dim=-2,
        )

        return G_EHp_6x3

    def get_G_Epm_3x6(self, r_probe, r_source, wavelength):
        G_EHp_6x3 = torch.cat(
            [
                self.get_G_Ep(r_probe, r_source, wavelength),
                self.get_G_Hp(r_probe, r_source, wavelength),
            ],
            dim=-2,
        )

        return G_EHp_6x3

    def get_G_Epm_3x6(self, r_probe, r_source, wavelength):
        G_Epm_3x6 = torch.cat(
            [
                self.get_G_Ep(r_probe, r_source, wavelength),
                self.get_G_Em(r_probe, r_source, wavelength),
            ],
            dim=-1,
        )

        return G_Epm_3x6

    def get_G_EHm_6x3(self, r_probe, r_source, wavelength):
        """magnetic field Green's tensor for a pair of p/m dipoles

        r_probe: field evaluation position, r_source: emitter-pair position
        """
        G_EHm_6x3 = torch.cat(
            [
                self.get_G_Em(r_probe, r_source, wavelength),
                self.get_G_Hm(r_probe, r_source, wavelength),
            ],
            dim=-2,
        )

        return G_EHm_6x3

    def get_G_Hpm_3x6(self, r_probe, r_source, wavelength):
        G_Hpm_3x6 = torch.cat(
            [
                self.get_G_Hp(r_probe, r_source, wavelength),
                self.get_G_Hm(r_probe, r_source, wavelength),
            ],
            dim=-1,
        )

        return G_Hpm_3x6

    # --- far-field approximation
    def get_G_Ep_farfield(self, r_probe, r_source, wavelength):
        """Electric field asymptotic far-field Green's tensor for an electric dipole

        r_probe: field evaluation position, r_source: emitter position
        """
        warnings.warn(
            "No `G_Ep` ff approximation implemented. Using full Green's tensor."
        )
        return self.get_G_Ep(r_probe, r_source, wavelength)

    def get_G_Em_farfield(self, r_probe, r_source, wavelength):
        """Electric field asymptotic far-field Green's tensor for a magnetic dipole

        r_probe: field evaluation position, r_source: emitter position
        """
        warnings.warn(
            "No `G_Em` ff approximation implemented. Using full Green's tensor."
        )
        return self.get_G_Em(r_probe, r_source, wavelength)


# --- illumination field defining base class
class IlluminationfieldBase:
    """base class defining the illumination E/H field"""

    __name__ = "illumination base class"

    def __init__(
        self,
        device: torch.device = None,
    ):
        """_summary_

        Args:
            device (torch.device, optional): Defaults to 'cpu'.
        """
        if device is None:
            self.device = get_default_device()
        else:
            self.device = device

        self.n_dim = -1  # problem dimension needs to be set by child class

    def __repr__(self, verbose=False):
        """description of illumination field"""
        out_str = " ------ {} -------".format(self.__name__)
        return out_str

    def copy(self):
        import copy

        return copy.deepcopy(self)

    def get_info(self):
        return dict(name=self.__name__)

    def set_device(self, device):
        """move all tensors of the class to device"""
        self.device = device

    def get_efield(
        self, r_probe: torch.Tensor, wavelength: float, environment: EnvironmentBase
    ):
        """evaluate illumination electric field at position(s) `r_probe`"""
        raise NotImplementedError(
            "electric field evaluation needs to be implemented in child class."
        )

    def get_hfield(
        self, r_probe: torch.Tensor, wavelength: float, environment: EnvironmentBase
    ):
        """evaluate illumination magnetic field at position(s) `r_probe`"""
        raise NotImplementedError(
            "magnetic field evaluation needs to be implemented in child class."
        )

    def get_field(self, r_probe, wavelength: float, environment: EnvironmentBase):
        """return electro-magnetic fields at `r_probe` as `Field` object"""
        from torchgdm.field import Field

        if type(r_probe) == dict:
            _r = r_probe["r_probe"]
        else:
            _r = r_probe

        e0 = self.get_efield(r_probe=_r, wavelength=wavelength, environment=environment)
        h0 = self.get_hfield(r_probe=_r, wavelength=wavelength, environment=environment)

        return Field(
            positions=r_probe,
            efield=e0.unsqueeze(0),
            hfield=h0.unsqueeze(0),
            device=self.device,
        )



# encoding=utf-8
"""contains collections of Green's tensors,
describing environments and their boundary conditions
"""
from . import inc_fields

from .dyads import EnvHomogeneous3D

from .inc_fields import NullField
from .inc_fields import PlaneWave
from .inc_fields import GaussianParaxial
from .inc_fields import ElectricDipole
from .inc_fields import MagneticDipole


# encoding=utf-8
# %%
import unittest

import torch

import torchgdm as tg


class TestSimCrossSections(unittest.TestCase):

    def setUp(self):
        self.verbose = False
        if self.verbose:
            print("testing cross section calculation with discretized simulation...")

        # --- determine if GPU is available
        self.devices = ["cpu"]
        if torch.cuda.is_available():
            self.devices.append("cuda:0")

        # --- setup a test case
        # - environment
        eps_env = 1.33
        mat_env = tg.materials.MatConstant(eps_env)
        env = tg.env.freespace_3d.EnvHomogeneous3D(env_material=mat_env)

        # - illumination field(s)
        wavelengths = torch.linspace(500, 800, 5)
        e_inc_list = [
            tg.env.freespace_3d.PlaneWave(e0p=1.0, e0s=0.6, inc_angle=torch.pi),
        ]

        # - structure
        step = 20.0
        mat_struct = tg.materials.MatConstant(eps=7.0 + 0.15j)
        struct1 = tg.struct.StructDiscretizedCubic3D(
            tg.struct.volume.cuboid(l=2, w=4, h=3), step, mat_struct
        )

        self.sim = tg.simulation.Simulation(
            structures=[struct1],
            environment=env,
            illumination_fields=e_inc_list,
            wavelengths=wavelengths,
        )

    def test_calculate_cs_spectra(self):
        for device in self.devices:
            self.sim.set_device(device)
            self.sim.run(verbose=False, progress_bar=False)

            # calculate spectra and compare with reference

            # via private API
            spec_ext_private = tg.tools.batch.calc_spectrum(
                self.sim, tg.postproc.crosssect.ecs, progress_bar=False
            )["ecs"].squeeze()
            spec_abs_private = tg.tools.batch.calc_spectrum(
                self.sim, tg.postproc.crosssect.acs, progress_bar=False
            )["acs"].squeeze()

            # via public API
            res_full = tg.tools.batch.calc_spectrum(
                self.sim, tg.postproc.crosssect.total, progress_bar=False
            )
            spec_scs = res_full["scs"].squeeze()
            spec_ext = res_full["ecs"].squeeze()
            spec_abs = res_full["acs"].squeeze()

            # print(spec_ext)
            # print(spec_scs)
            # print(spec_abs)

            # compare to pre-calculated reference
            spec_ext_truth = torch.as_tensor(
               [453.8657837, 279.2537231, 187.5266266, 134.9423523, 102.5867004],
                dtype=tg.constants.DTYPE_FLOAT,
                device=device,
            )
            spec_scs_truth = torch.as_tensor(
                [380.5903320, 213.5857544, 128.9793243,  82.4661942,  55.1847038],
                dtype=tg.constants.DTYPE_FLOAT,
                device=device,
            )
            spec_abs_truth = torch.as_tensor(
                [73.2754517, 65.6679688, 58.5473061, 52.4761581, 47.4019966],
                dtype=tg.constants.DTYPE_FLOAT,
                device=device,
            )

            torch.testing.assert_close(spec_ext, spec_ext_truth, rtol=1e-3, atol=1e-3)
            torch.testing.assert_close(spec_scs, spec_scs_truth, rtol=1e-3, atol=1e-3)
            torch.testing.assert_close(spec_abs, spec_abs_truth, rtol=1e-3, atol=1e-3)

            # the individual functions must yield same result as the combined `scattering` function
            torch.testing.assert_close(spec_ext_private, spec_ext)
            torch.testing.assert_close(spec_abs_private, spec_abs)

            if self.verbose:
                print(
                    "  - {}: cross-section spectra (extinct./abs./scat.) test passed.".format(
                        device
                    )
                )


# %%

# %%
if __name__ == "__main__":
    print("testing cross section calculation...")
    torch.set_printoptions(precision=7)
    unittest.main(argv=["first-arg-is-ignored"], exit=False)



# -*- coding: utf-8 -*-
"""manage installation"""
from setuptools import setup, find_namespace_packages, Extension
import os
import re


# =============================================================================
# helper functions to extract meta-info from package
# =============================================================================
def read_version_file(*parts):
    return open(os.path.join(*parts), "r").read()


def read(fname):
    return open(os.path.join(os.path.dirname(__file__), fname)).read()


def find_version(*file_paths):
    version_file = read_version_file(*file_paths)
    version_match = re.search(r"^__version__ = ['\"]([^'\"]*)['\"]", version_file, re.M)
    if version_match:
        return version_match.group(1)
    raise RuntimeError("Unable to find version string.")


def find_name(*file_paths):
    version_file = read_version_file(*file_paths)
    version_match = re.search(r"^__name__ = ['\"]([^'\"]*)['\"]", version_file, re.M)
    if version_match:
        return version_match.group(1)
    raise RuntimeError("Unable to find name string.")


def find_author(*file_paths):
    version_file = read_version_file(*file_paths)
    version_match = re.search(r"^__author__ = ['\"]([^'\"]*)['\"]", version_file, re.M)
    if version_match:
        return version_match.group(1)
    raise RuntimeError("Unable to find author string.")


# =============================================================================
# package module list
# =============================================================================
package_list = find_namespace_packages(where=".", include=["torchgdm*"])


# =============================================================================
# main setup
# =============================================================================
setup(
    name=find_name("torchgdm", "__init__.py"),
    version=find_version("torchgdm", "__init__.py"),
    author=find_author("torchgdm", "__init__.py"),
    author_email="pwiecha@laas.fr",
    description=(
        "A python electrodynamical CPU/GPU solver written in PyTorch,"
        "With full support of PyTorch's automatic differentiation."
        "Based on the Green's Dyadic Method (GDM, volume integral technique "
        "in frequency domain)."
    ),
    license="GPLv3+",
    long_description=read("README.md"),
    long_description_content_type="text/markdown",
    packages=package_list,
    package_data={"torchgdm.materials.data": ["*.yml"]},
    classifiers=[
        "Development Status :: 5 - Production/Stable",
        "Topic :: Scientific/Engineering :: Physics",
        "Environment :: Console",
        "Programming Language :: Python :: 3.9",
        "Programming Language :: Python :: 3.10",
        "Programming Language :: Python :: 3.11",
        "Programming Language :: Python :: 3.12",
        "License :: OSI Approved :: GNU General Public License v3 or later (GPLv3+)",
        "Intended Audience :: Science/Research",
    ],
    url="https://gitlab.com/wiechapeter/torchgdm",
    download_url="",
    keywords=[
        "volume integral method",
        "green dyadic method",
        "coupled dipoles method",
        "effective polarizabilities",
        "electrodynamical simulations",
        "nano optics",
        "frequency-domain",
        "automatic differentiation",
    ],
    install_requires=["torch>=2.0.0"],
    python_requires=">=3.9",
)



# -*- coding: utf-8 -*-
"""package for torchgdm structures
"""
from . import volume
from . import point
from . import surface_2d
from . import line_2d

# - 3D
# from .volume.pola import StructDiscretized3D
from .volume.pola import StructDiscretizedCubic3D
from .volume.pola import StructDiscretizedHexagonal3D

from .point.pola import StructEffPola3D
from .point.pola import StructMieSphereEffPola3D  # Mie core-shell

# - 2D
# from .surface_2d.pola import StructDiscretized2D
from .surface_2d.pola import StructDiscretizedSquare2D

from .line_2d.pola import StructEffPola2D


# encoding=utf-8
# %%
import unittest

import torch

import torchgdm as tg


class TestGreensFreeSpace2D(unittest.TestCase):

    def setUp(self):
        self.verbose = False

        # --- determine if GPU is available
        self.devices = ["cpu"]
        if torch.cuda.is_available():
            self.devices.append("cuda")

        # --- setup a test case
        eps_env = 1.0
        self.mat_env = tg.materials.MatConstant(eps_env)

    def test_G2D_ky0(self):
        for device in self.devices:
            # positions and wavelength in nm
            env = tg.env.freespace_2d.EnvHomogeneous2D(env_material=self.mat_env)

            wl = torch.as_tensor(730.0)
            r_probe = torch.as_tensor(
                [50, 0, 70], dtype=tg.constants.DTYPE_FLOAT, device=device
            )
            r_source = torch.as_tensor(
                [73, 0, 99], dtype=tg.constants.DTYPE_FLOAT, device=device
            )

            G_ref = torch.as_tensor(
                [
                    [
                        [
                            -2.5220914e-04 + 1.1310077e-04j,
                            0.0000000e00 + 0.0000000e00j,
                            1.4593793e-03 + 1.4253797e-06j,
                            -0.0000000e00 + 0.0000000e00j,
                            2.8679229e-05 - 3.9634385e-04j,
                            0.0000000e00 - 0.0000000e00j,
                        ],
                        [
                            0.0000000e00 + 0.0000000e00j,
                            1.7823000e-04 + 2.2686829e-04j,
                            0.0000000e00 + 0.0000000e00j,
                            -2.8679229e-05 + 3.9634385e-04j,
                            -0.0000000e00 + 0.0000000e00j,
                            2.2745597e-05 - 3.1434168e-04j,
                        ],
                        [
                            1.4593793e-03 + 1.4253797e-06j,
                            0.0000000e00 + 0.0000000e00j,
                            4.3043910e-04 + 1.1376752e-04j,
                            0.0000000e00 + 0.0000000e00j,
                            -2.2745597e-05 + 3.1434168e-04j,
                            -0.0000000e00 + 0.0000000e00j,
                        ],
                        [
                            0.0000000e00 + 0.0000000e00j,
                            -2.8679229e-05 + 3.9634385e-04j,
                            -0.0000000e00 + 0.0000000e00j,
                            -2.5220914e-04 + 1.1310077e-04j,
                            0.0000000e00 + 0.0000000e00j,
                            1.4593793e-03 + 1.4253797e-06j,
                        ],
                        [
                            2.8679229e-05 - 3.9634385e-04j,
                            0.0000000e00 + 0.0000000e00j,
                            -2.2745597e-05 + 3.1434168e-04j,
                            0.0000000e00 + 0.0000000e00j,
                            1.7823000e-04 + 2.2686829e-04j,
                            0.0000000e00 + 0.0000000e00j,
                        ],
                        [
                            0.0000000e00 - 0.0000000e00j,
                            2.2745597e-05 - 3.1434168e-04j,
                            0.0000000e00 + 0.0000000e00j,
                            1.4593793e-03 + 1.4253797e-06j,
                            0.0000000e00 + 0.0000000e00j,
                            4.3043910e-04 + 1.1376752e-04j,
                        ],
                    ]
                ],
                device=device,
                dtype=tg.constants.DTYPE_COMPLEX,
            )
            G_test = env.get_G_6x6(r_probe, r_source, wl)
            # torch.set_printoptions(precision=7)
            # print(G_test)

            torch.testing.assert_close(G_test, G_ref)
            if self.verbose:
                print(
                    "  - {}: G-2D vacuum normal incidence test passed.".format(device)
                )

    def test_G2D_ky(self):
        for device in self.devices:
            # positions and wavelength in nm
            env = tg.env.freespace_2d.EnvHomogeneous2D(
                env_material=self.mat_env, inc_angle_y=-torch.pi / 4
            )

            wl = torch.as_tensor(530.0)
            r_probe = torch.as_tensor(
                [20, 0, 33], dtype=tg.constants.DTYPE_FLOAT, device=device
            )
            r_source = torch.as_tensor(
                [-65, 0, -5], dtype=tg.constants.DTYPE_FLOAT, device=device
            )

            G_ref = torch.as_tensor(
                [
                    [
                        [
                            2.1607113e-04 + 2.8792763e-04j,
                            -7.2812261e-05 + 2.0167719e-04j,
                            2.0219064e-04 + 5.9515496e-06j,
                            -0.0000000e00 + 0.0000000e00j,
                            -4.6034576e-05 + 1.2750768e-04j,
                            3.3123157e-05 + 2.6643940e-04j,
                        ],
                        [
                            -7.2812261e-05 + 2.0167719e-04j,
                            2.3421604e-05 + 1.8840107e-04j,
                            -3.2551361e-05 + 9.0161549e-05j,
                            4.6034576e-05 - 1.2750768e-04j,
                            -0.0000000e00 + 0.0000000e00j,
                            -1.0297208e-04 + 2.8521454e-04j,
                        ],
                        [
                            2.0219064e-04 + 5.9515496e-06j,
                            -3.2551361e-05 + 9.0161549e-05j,
                            -1.4580631e-04 + 2.7727569e-04j,
                            -3.3123157e-05 - 2.6643940e-04j,
                            1.0297208e-04 - 2.8521454e-04j,
                            -0.0000000e00 + 0.0000000e00j,
                        ],
                        [
                            0.0000000e00 + 0.0000000e00j,
                            4.6034576e-05 - 1.2750768e-04j,
                            -3.3123157e-05 - 2.6643940e-04j,
                            2.1607113e-04 + 2.8792763e-04j,
                            -7.2812261e-05 + 2.0167719e-04j,
                            2.0219064e-04 + 5.9515496e-06j,
                        ],
                        [
                            -4.6034576e-05 + 1.2750768e-04j,
                            0.0000000e00 + 0.0000000e00j,
                            1.0297208e-04 - 2.8521454e-04j,
                            -7.2812261e-05 + 2.0167719e-04j,
                            2.3421604e-05 + 1.8840107e-04j,
                            -3.2551361e-05 + 9.0161549e-05j,
                        ],
                        [
                            3.3123157e-05 + 2.6643940e-04j,
                            -1.0297208e-04 + 2.8521454e-04j,
                            0.0000000e00 + 0.0000000e00j,
                            2.0219064e-04 + 5.9515496e-06j,
                            -3.2551361e-05 + 9.0161549e-05j,
                            -1.4580631e-04 + 2.7727569e-04j,
                        ],
                    ]
                ],
                device=device,
                dtype=tg.constants.DTYPE_COMPLEX,
            )
            G_test = env.get_G_6x6(r_probe, r_source, wl)
            # torch.set_printoptions(precision=7)
            # print(G_test)

            torch.testing.assert_close(G_test, G_ref)
            if self.verbose:
                print(
                    "  - {}: G-2D vacuum normal incidence test passed.".format(device)
                )

    def test_subtensors(self):
        for device in self.devices:
            for angle_ky in [0, torch.pi / 4, torch.pi / 2.1]:
                env = tg.env.freespace_2d.EnvHomogeneous2D(
                    env_material=self.mat_env, inc_angle_y=angle_ky
                )
                # positions and wavelength in nm
                wl = torch.as_tensor(
                    730.0, dtype=tg.constants.DTYPE_FLOAT, device=device
                )
                r_probe = torch.as_tensor(
                    [250, 0, 5], dtype=tg.constants.DTYPE_FLOAT, device=device
                )
                r_source = torch.as_tensor(
                    [244, 0, -2], dtype=tg.constants.DTYPE_FLOAT, device=device
                )
                env.set_device(device)

                G = env.get_G_6x6(r_probe, r_source, wl)
                GEp = env.get_G_Ep(r_probe, r_source, wl)
                GEm = env.get_G_Em(r_probe, r_source, wl)
                GHp = env.get_G_Hp(r_probe, r_source, wl)
                GHm = env.get_G_Hm(r_probe, r_source, wl)
                GEHp = env.get_G_EHp_6x3(r_probe, r_source, wl)
                GEHm = env.get_G_EHm_6x3(r_probe, r_source, wl)
                GEpm = env.get_G_Epm_3x6(r_probe, r_source, wl)
                GHpm = env.get_G_Hpm_3x6(r_probe, r_source, wl)

                # test correlated 3x3 tensors
                torch.testing.assert_close(GEp[0], GHm[0], rtol=1e-9, atol=1e-9)
                torch.testing.assert_close(GEm[0], -1 * GHp[0], rtol=1e-9, atol=1e-9)

                # test 3x3 tensors in 6x6 tensor
                torch.testing.assert_close(GEp[0], G[0, :3, :3], rtol=1e-9, atol=1e-9)
                torch.testing.assert_close(GHm[0], G[0, 3:, 3:], rtol=1e-9, atol=1e-9)
                torch.testing.assert_close(GHp[0], G[0, 3:, :3], rtol=1e-9, atol=1e-9)
                torch.testing.assert_close(GEm[0], G[0, :3, 3:], rtol=1e-9, atol=1e-9)

                # test 6x3 and 3x6 tensors
                torch.testing.assert_close(GEHp[0], G[0, :, :3], rtol=1e-9, atol=1e-9)
                torch.testing.assert_close(GEHm[0], G[0, :, 3:], rtol=1e-9, atol=1e-9)
                torch.testing.assert_close(GEpm[0], G[0, :3, :], rtol=1e-9, atol=1e-9)
                torch.testing.assert_close(GHpm[0], G[0, 3:, :], rtol=1e-9, atol=1e-9)

                if self.verbose:
                    print(
                        "  - {}: 2D subtensor consistency test passed.".format(device)
                    )


class TestGreensFreeSpace3D(unittest.TestCase):

    def setUp(self):
        self.verbose = False

        # --- determine if GPU is available
        self.devices = ["cpu"]
        if torch.cuda.is_available():
            self.devices.append("cuda:0")

        # --- setup a test case
        eps_env = 1.0
        mat_env = tg.materials.MatConstant(eps_env)
        self.env = tg.env.freespace_3d.EnvHomogeneous3D(env_material=mat_env)

    def test_G(self):
        for device in self.devices:
            # positions and wavelength in nm
            wl = torch.as_tensor(730.0)
            r_probe = torch.as_tensor(
                [50, 60, 70], dtype=tg.constants.DTYPE_FLOAT, device=device
            )
            r_source = torch.as_tensor(
                [73, -45, 99], dtype=tg.constants.DTYPE_FLOAT, device=device
            )

            G_ref = torch.as_tensor(
                [
                    [
                        [
                            -4.9245153e-07 + 3.5233521e-07j,
                            -5.0107946e-07 - 7.1189419e-09j,
                            1.3839337e-07 + 1.9661783e-09j,
                            0.0000000e00 + 0.0000000e00j,
                            4.8337917e-08 - 2.4580962e-07j,
                            1.7501657e-07 - 8.9000031e-07j,
                        ],
                        [
                            -5.0107946e-07 - 7.1189419e-09j,
                            1.6853248e-06 + 3.8327528e-07j,
                            -6.3179590e-07 - 8.9760874e-09j,
                            -4.8337917e-08 + 2.4580962e-07j,
                            0.0000000e00 + 0.0000000e00j,
                            3.8336964e-08 - 1.9495246e-07j,
                        ],
                        [
                            1.3839337e-07 + 1.9661783e-09j,
                            -6.3179590e-07 - 8.9760874e-09j,
                            -4.2771575e-07 + 3.5325488e-07j,
                            -1.7501657e-07 + 8.9000031e-07j,
                            -3.8336964e-08 + 1.9495246e-07j,
                            0.0000000e00 + 0.0000000e00j,
                        ],
                        [
                            0.0000000e00 - 0.0000000e00j,
                            -4.8337917e-08 + 2.4580962e-07j,
                            -1.7501657e-07 + 8.9000031e-07j,
                            -4.9245153e-07 + 3.5233521e-07j,
                            -5.0107946e-07 - 7.1189419e-09j,
                            1.3839337e-07 + 1.9661783e-09j,
                        ],
                        [
                            4.8337917e-08 - 2.4580962e-07j,
                            0.0000000e00 - 0.0000000e00j,
                            -3.8336964e-08 + 1.9495246e-07j,
                            -5.0107946e-07 - 7.1189419e-09j,
                            1.6853248e-06 + 3.8327528e-07j,
                            -6.3179590e-07 - 8.9760874e-09j,
                        ],
                        [
                            1.7501657e-07 - 8.9000031e-07j,
                            3.8336964e-08 - 1.9495246e-07j,
                            0.0000000e00 - 0.0000000e00j,
                            1.3839337e-07 + 1.9661783e-09j,
                            -6.3179590e-07 - 8.9760874e-09j,
                            -4.2771575e-07 + 3.5325488e-07j,
                        ],
                    ]
                ],
                device=device,
                dtype=tg.constants.DTYPE_COMPLEX,
            )
            G_test = self.env.get_G_6x6(r_probe, r_source, wl)
            # torch.set_printoptions(precision=7); print(G_test)

            torch.testing.assert_close(G_test, G_ref)
            if self.verbose:
                print("  - {}: 3D vacuum test passed.".format(device))

    def test_subtensors(self):
        for device in self.devices:
            # positions and wavelength in nm
            wl = torch.as_tensor(730.0, dtype=tg.constants.DTYPE_FLOAT, device=device)
            r_probe = torch.as_tensor(
                [250, 10, 5], dtype=tg.constants.DTYPE_FLOAT, device=device
            )
            r_source = torch.as_tensor(
                [244, -5, -2], dtype=tg.constants.DTYPE_FLOAT, device=device
            )
            self.env.set_device(device)

            G = self.env.get_G_6x6(r_probe, r_source, wl)
            GEp = self.env.get_G_Ep(r_probe, r_source, wl)
            GEm = self.env.get_G_Em(r_probe, r_source, wl)
            GHp = self.env.get_G_Hp(r_probe, r_source, wl)
            GHm = self.env.get_G_Hm(r_probe, r_source, wl)
            GEHp = self.env.get_G_EHp_6x3(r_probe, r_source, wl)
            GEHm = self.env.get_G_EHm_6x3(r_probe, r_source, wl)
            GEpm = self.env.get_G_Epm_3x6(r_probe, r_source, wl)
            GHpm = self.env.get_G_Hpm_3x6(r_probe, r_source, wl)

            # test correlated 3x3 tensors
            torch.testing.assert_close(GEp[0], GHm[0], rtol=1e-9, atol=1e-9)
            torch.testing.assert_close(GEm[0], -1 * GHp[0], rtol=1e-9, atol=1e-9)

            # test 3x3 tensors in 6x6 tensor
            torch.testing.assert_close(GEp[0], G[0, :3, :3], rtol=1e-9, atol=1e-9)
            torch.testing.assert_close(GHm[0], G[0, 3:, 3:], rtol=1e-9, atol=1e-9)
            torch.testing.assert_close(GHp[0], G[0, 3:, :3], rtol=1e-9, atol=1e-9)
            torch.testing.assert_close(GEm[0], G[0, :3, 3:], rtol=1e-9, atol=1e-9)

            # test 6x3 and 3x6 tensors
            torch.testing.assert_close(GEHp[0], G[0, :, :3], rtol=1e-9, atol=1e-9)
            torch.testing.assert_close(GEHm[0], G[0, :, 3:], rtol=1e-9, atol=1e-9)
            torch.testing.assert_close(GEpm[0], G[0, :3, :], rtol=1e-9, atol=1e-9)
            torch.testing.assert_close(GHpm[0], G[0, 3:, :], rtol=1e-9, atol=1e-9)

            if self.verbose:
                print("  - {}: 3D subtensor consistency test passed.".format(device))


# %%
if __name__ == "__main__":
    print("testing free space (2d & 3d) Green's tensors...")
    torch.set_printoptions(precision=7)
    unittest.main(argv=["first-arg-is-ignored"], exit=False)



# -*- coding: utf-8 -*-
"""
torchgdm Green's function and LDOS tools
"""
# %%
import warnings

import torch

import torchgdm as tg
from torchgdm.constants import DTYPE_FLOAT, DTYPE_COMPLEX
from torchgdm.tools.misc import get_default_device
from torchgdm.linearsystem import _expand_dimensions
from torchgdm.tools.misc import test_positional_input
from torchgdm.tools.batch import batched


def G(
    sim,
    r_probe: torch.Tensor,
    r_source: torch.Tensor,
    wavelength: float,
    use_sim_copy=True,
    progress_bar=True,
    verbose=0,
    **kwargs,
):
    """calculate the Green's tensor of the complex environment defined by `sim`

    Args:
        sim (`torchgdm.Simulation`): simulation
        r_probe (torch.Tensor): probe position(s)
        r_source (torch.Tensor): source location(s)
        wavelength (float): in nm
        use_sim_copy (bool, optional): Use copy of simulation or use simulation in-place. May not work in some autograd scenario. Defaults to True.
        progress_bar (bool, optional): Show progress bar. Defaults to True.
        verbose (bool, optional): Print status of underlying simulation call. Defaults to False.

    Returns:
        dict: contains Green's tensors for all source/probe position combinations:
            - "G_6x6": full 6x6 tensor
            - "G_Ep": electric-electric 3x3 tensor
            - "G_Em": electric-magnetic 3x3 tensor
            - "G_Hp": magnetic-electric 3x3 tensor
            - "G_Hm": magnetic-magnetic 3x3 tensor
    """
    if type(r_probe) == dict:
        r_probe = r_probe["r_probe"]
    if type(r_source) == dict:
        r_source = r_source["r_probe"]

    if r_probe is None:
        r_probe = r_source

    r_probe = torch.as_tensor(r_probe, dtype=DTYPE_FLOAT, device=sim.device)
    r_source = torch.as_tensor(r_source, dtype=DTYPE_FLOAT, device=sim.device)
    r_probe, r_source = test_positional_input(r_probe, r_source)

    if len(r_source.shape) == 1:
        r_source = r_source.unsqueeze(0)
    if len(r_probe.shape) == 1:
        r_probe = r_probe.unsqueeze(0)

    if use_sim_copy:
        _sim = sim.copy()
    else:
        _sim = sim
        warnings.warn("Using simulation in-place modifies the illuminations!")

    # --- configure dipole illuminations with perpendicular orientations
    e_inc_list = []
    for r_dp in r_source:
        # electric illumination
        for p_dp in [[1.0, 0, 0], [0, 1.0, 0], [0, 0, 1.0]]:
            e_inc_list.append(
                tg.env.freespace_3d.ElectricDipole(
                    r_source=r_dp, p_source=p_dp, device=sim.device
                )
            )
        # magnetic illumination
        for p_dp in [[1.0, 0, 0], [0, 1.0, 0], [0, 0, 1.0]]:
            e_inc_list.append(
                tg.env.freespace_3d.MagneticDipole(
                    r_source=r_dp, m_source=p_dp, device=sim.device
                )
            )
    _sim.illumination_fields = e_inc_list

    # run the simulation
    _sim.run(calc_missing=False, verbose=verbose, progress_bar=progress_bar)

    nf_results = tg.postproc.fields.nf(
        _sim,
        wavelength,
        r_probe=r_probe,
        progress_bar=progress_bar,
        pgr_bar_title="Green's tensor batch evaluation",
        **kwargs,
    )
    e_sca = nf_results["tot"]

    G_E = e_sca.efield
    G_H = e_sca.hfield

    # reshape. second last column e-field components, last column dipole orientations
    G_E = G_E.reshape(-1, 6, *G_E.shape[1:])
    G_E = torch.moveaxis(G_E, 1, -1)
    G_H = G_H.reshape(-1, 6, *G_H.shape[1:])
    G_H = torch.moveaxis(G_H, 1, -1)

    G_Ep = G_E[..., :3]
    G_Em = G_E[..., 3:]
    G_Hp = G_H[..., :3]
    G_Hm = G_H[..., 3:]

    G_6x6 = torch.cat([G_E, G_H], dim=-2)

    return dict(G_6x6=G_6x6, G_Ep=G_Ep, G_Em=G_Em, G_Hp=G_Hp, G_Hm=G_Hm)


def ldos(
    sim,
    r_probe: torch.Tensor,
    wavelength: float,
    progress_bar=True,
    verbose=0,
    **kwargs,
):
    """calculate the relative LDOS of the complex environment defined by `sim`

    this is yields the same results as Green's tensor evaluation, but
    is optimized for evaluation of (r_source, r_source) tuples

    Args:
        sim (`torchgdm.Simulation`): simulation
        r_probe (torch.Tensor): probe position(s)
        wavelength (float): in nm
        progress_bar (bool, optional): Show progress bar. Defaults to True.
        verbose (bool, optional): Print status info. Defaults to False.

    Returns:
        dict: contains Green's tensors, partial and averaged LDOS all probe positions:
            - "G_ii": full 6x6 tensor
            - "ldos_partial": partial electric (first 3) and magnetic (last 3) LDOS. (Diagonal elements of Green's tensor)
            - "ldos_e": full electric LDOS
            - "ldos_m": full magnetic LDOS
    """
    if verbose:
        print("-" * 60)
        print(sim)
        print("\nrun LDOS simulation.")

    # - prep
    if type(r_probe) == dict:
        r_src = r_probe["r_probe"]
    else:
        r_src = r_probe

    r_src = torch.as_tensor(r_src, dtype=DTYPE_FLOAT, device=sim.device)

    if len(r_src.shape) == 1:
        r_src = r_src.unsqueeze(0)

    r_geo = sim.get_all_positions()

    # ---
    # evaluate and integrate Q.alpha.K.S
    # --- dyads
    # TODO: memory efficiency
    #   - use polarizable positions only
    #   - Q and S as reduced coupling matrices
    if verbose:
        print("get alphas...", end="")
    alpha = sim._get_all_polarizabilitites_6x6(wavelength)

    # Note: this is not memory / computational efficient!
    if verbose:
        print(" Done.\nsolve linear system...", end="")
    full_linsys = tg.linearsystem.LinearSystemFullInverse(device=sim.device)
    K_red = full_linsys.get_generalized_propagator(sim, wavelength=wavelength)
    K = _expand_dimensions(K_red, N=6)
    tg.tools.misc.purge_mem(K_red)
    if verbose:
        print(" Done.")

    @batched(
        batch_kwarg="r_src", arg_position=2, out_dim_batch=0, default_batch_size=128
    )
    def _eval_Gii(sim, wavelength, r_src, r_geo, K):
        # dyads: source -> structure
        Q = sim.environment.get_G_6x6(
            r_geo.unsqueeze(0), r_src.unsqueeze(1), wavelength
        )

        # dyads: structure -> probe
        S = sim.environment.get_G_6x6(
            r_src.unsqueeze(0), r_geo.unsqueeze(1), wavelength
        )

        # calculate double integral of tensor products:
        # sum_j sum_k (Q_nj * ((alpha_j * K_jk) * S_kn))
        aK = torch.einsum("jab, jkbc -> jkac", [alpha, K])
        aKS = torch.einsum("jkab, kmbc -> jmac", [aK, S])
        tg.tools.misc.purge_mem(aK)
        QaKS = torch.einsum("njab, jnbc -> nca", [Q, aKS])
        tg.tools.misc.purge_mem(aKS)

        return QaKS

    G_ii = _eval_Gii(
        sim,
        wavelength,
        r_src=r_src,
        r_geo=r_geo,
        K=K,
        progress_bar=progress_bar,
        pgr_bar_title="LDOS batch evaluation",
        **kwargs,
    )

    k0 = 2 * torch.pi / wavelength
    prefactor = (3.0 / 2.0) * (1.0 / k0**3)
    G_ii_diag = torch.diagonal(G_ii, offset=0, dim1=1, dim2=2)

    ldos_partial = 1.0 + prefactor * G_ii_diag.imag
    ldos_e = 1.0 + prefactor * torch.sum(G_ii_diag[:, :3].imag, dim=-1)
    ldos_m = 1.0 + prefactor * torch.sum(G_ii_diag[:, 3:].imag, dim=-1)

    return dict(G_ii=G_ii, ldos_partial=ldos_partial, ldos_e=ldos_e, ldos_m=ldos_m)


# - convenience wrapper
def G_6x6(
    sim,
    r_probe: torch.Tensor,
    r_source: torch.Tensor,
    wavelength: float,
    use_sim_copy=True,
    progress_bar=False,
    **kwargs,
):
    """calculate the Ep Green's tensor for `sim`

    wrapper to :func:`G`, see there for more documentation.

    Returns:
        torch.Tensor: full electric-magnetic Green's tensor for all source/probe position combinations
    """
    G_results = G(
        sim=sim,
        r_probe=r_probe,
        r_source=r_source,
        wavelength=wavelength,
        use_sim_copy=use_sim_copy,
        progress_bar=progress_bar,
        **kwargs,
    )
    return G_results["G_6x6"]


def G_Ep(
    sim,
    r_probe: torch.Tensor,
    r_source: torch.Tensor,
    wavelength: float,
    use_sim_copy=True,
    progress_bar=False,
    **kwargs,
):
    """calculate the Ep Green's tensor for `sim`

    wrapper to :func:`G`, see there for more documentation.

    Returns:
        torch.Tensor: electric-electric Green's tensor for all source/probe position combinations
    """
    G_results = G(
        sim=sim,
        r_probe=r_probe,
        r_source=r_source,
        wavelength=wavelength,
        use_sim_copy=use_sim_copy,
        progress_bar=progress_bar,
        **kwargs,
    )
    return G_results["G_Ep"]


def G_Em(
    sim,
    r_probe: torch.Tensor,
    r_source: torch.Tensor,
    wavelength: float,
    use_sim_copy=True,
    progress_bar=False,
    **kwargs,
):
    """calculate the Em Green's tensor for `sim`

    wrapper to :func:`G`, see there for more documentation.

    Returns:
        torch.Tensor: electric-magnetic Green's tensor for all source/probe position combinations
    """
    G_results = G(
        sim=sim,
        r_probe=r_probe,
        r_source=r_source,
        wavelength=wavelength,
        use_sim_copy=use_sim_copy,
        progress_bar=progress_bar,
        **kwargs,
    )
    return G_results["G_Em"]


def G_Hp(
    sim,
    r_probe: torch.Tensor,
    r_source: torch.Tensor,
    wavelength: float,
    use_sim_copy=True,
    progress_bar=False,
    **kwargs,
):
    """calculate the Hp Green's tensor for `sim`

    wrapper to :func:`G`, see there for more documentation.

    Returns:
        torch.Tensor: magnetic-electric Green's tensor for all source/probe position combinations
    """
    G_results = G(
        sim=sim,
        r_probe=r_probe,
        r_source=r_source,
        wavelength=wavelength,
        use_sim_copy=use_sim_copy,
        progress_bar=progress_bar,
        **kwargs,
    )
    return G_results["G_Hp"]


def G_Hm(
    sim,
    r_probe: torch.Tensor,
    r_source: torch.Tensor,
    wavelength: float,
    use_sim_copy=True,
    progress_bar=False,
    **kwargs,
):
    """calculate the Hm Green's tensor for `sim`

    wrapper to :func:`G`, see there for more documentation.

    Returns:
        torch.Tensor: magnetic-magnetic Green's tensor for all source/probe position combinations
    """
    G_results = G(
        sim=sim,
        r_probe=r_probe,
        r_source=r_source,
        wavelength=wavelength,
        use_sim_copy=use_sim_copy,
        progress_bar=progress_bar,
        **kwargs,
    )
    return G_results["G_Hm"]


def _G_direct(
    sim,
    r_probe: torch.Tensor,
    r_source: torch.Tensor,
    wavelength: float = None,
    **kwargs,
):
    """calculate the Green's tensor via direct evaluation

    should give the same results as :func:`G`, but significantly less memory efficient
    and incorrect inside a discretized nanostructure.

    Args:
        sim (`torchgdm.Simulation`): simulation
        r_probe (torch.Tensor): probe position(s)
        r_source (torch.Tensor): source location(s)
        wavelength (float): in nm
    """
    # - prep
    if wavelength is None:
        raise ValueError("Wavelength not given. Needs to be defined.")

    if type(r_probe) == dict:
        r_prb = r_probe["r_probe"]

    r_prb = torch.as_tensor(r_prb, dtype=DTYPE_FLOAT, device=sim.device)
    r_src = torch.as_tensor(r_source, dtype=DTYPE_FLOAT, device=sim.device)

    if len(r_prb.shape) == 1:
        r_prb = r_prb.unsqueeze(0)
    if len(r_src.shape) == 1:
        r_src = r_src.unsqueeze(0)

    r_geo = sim.get_all_positions()

    # ---
    # evaluate and integrate Q.alpha.K.S
    # --- dyads
    # TODO: memory efficiency
    #   - use polarizable positions only
    #   - Q and S as reduced coupling matrices
    alpha = sim._get_all_polarizabilitites_6x6(wavelength)

    # Note: this is not memory / computational efficient!
    full_linsys = tg.linearsystem.LinearSystemFullInverse(device=sim.device)
    K_red = full_linsys.get_generalized_propagator(sim, wavelength=wavelength)
    K = _expand_dimensions(K_red, N=6)
    tg.tools.misc.purge_mem(K_red)

    from torchgdm.tools.batch import batched

    @batched(
        batch_kwarg="r_src", arg_position=2, out_dim_batch=0, default_batch_size=1024
    )
    def _eval_Gii(sim, wavelength, r_src, r_geo, r_prb, K):
        # dyads: source -> structure
        Q = sim.environment.get_G_6x6(
            r_geo.unsqueeze(0), r_src.unsqueeze(1), wavelength
        )

        # dyads: structure -> probe
        S = sim.environment.get_G_6x6(
            r_prb.unsqueeze(0), r_geo.unsqueeze(1), wavelength
        )

        # dyads: source -> probe
        D = sim.environment.get_G_6x6(
            r_prb.unsqueeze(0), r_src.unsqueeze(1), wavelength
        )

        # calculate double integral of tensor products:
        # sum_j sum_k (Q_nj * ((alpha_j * K_jk) * S_kn))
        aK = torch.einsum("jab, jkbc -> jkac", [alpha, K])
        aKS = torch.einsum("jkab, kmbc -> jmac", [aK, S])
        tg.tools.misc.purge_mem(aK)
        QaKS = torch.einsum("mjab, jnbc -> mnca", [Q, aKS])
        tg.tools.misc.purge_mem(aKS)

        # return scattered + direct field
        return QaKS + D

    G = _eval_Gii(sim, wavelength, r_src=r_src, r_geo=r_geo, r_prb=r_prb, K=K, **kwargs)

    return G


# %% testing
if __name__ == "__main__":
    # %% TODO:
    # tests for Green's tensor vs vacuum in absence of structure
    # test for Green's tensor reciprocity
    # test for Green's tensor vs LDOS (for same result)

    import matplotlib.pyplot as plt

    device = "cpu"
    device = "cuda"

    # --- setup test case simulation
    # - illumination field(s)
    wavelengths = torch.tensor([550.0])

    # - environment
    eps_env = 1.0
    mat_env = tg.materials.MatConstant(eps_env)
    env = tg.env.freespace_3d.EnvHomogeneous3D(env_material=mat_env)

    # - dummy illumination field
    e_inc_dummy = tg.env.freespace_3d.NullField()

    # - first structure: volume discretization
    l = 20
    w = 10
    h = 5
    step = 20.0
    l = w = h = 7
    step = 22.0
    mat_struct = tg.materials.MatConstant(eps=14.0)

    struct_mesh = tg.struct.StructDiscretizedCubic3D(
        tg.struct.volume.cuboid(l, w, h), step, mat_struct
    )
    struct_mesh.set_center_of_mass([0, 0, 0])
    # struct_mesh += [100000, 100000, -400]
    struct_mesh.plot()

    # - full simulation
    sim = tg.simulation.Simulation(
        structures=[struct_mesh],
        environment=env,
        illumination_fields=[e_inc_dummy],
        wavelengths=wavelengths,
        device=device,
    )

    print("nr of dipoles: {}".format(len(sim.get_all_positions())))

    wavelength = wavelengths[0]
    r_source = [0, 0, 200]
    r_source = [[-220, 0, 0], [50, 0, 200]]
    r_probe = [[0, 0, 200], [100, 200, 300]]
    r_probe = [[100, 20, 200]]

    r_probe = tg.tools.geometry.coordinate_map_2d_square(500, 100, r3=100)
    # r_source = r_probe

    # r_probe = None

    G_results = G(
        sim,
        wavelength=wavelength,
        r_probe=r_probe,
        r_source=r_source,
        progress_bar=True,
    )

    G_6x6 = G_results["G_6x6"]
    G_Ep = G_results["G_Ep"]
    G_Hm = G_results["G_Hm"]

    tg.visu.visu2d.scalar2d._scalarfield(
        field_scalars=G_Ep[0, :, 0, 0].real, positions=r_probe["r_probe"]
    )  # %%
    # %% LDOS test
    # G_selfEp = G_Ep[torch.arange(G_Ep.shape[0]), torch.arange(G_Ep.shape[1]), ...]
    # G_selfHm = G_Hm[torch.arange(G_Hm.shape[0]), torch.arange(G_Hm.shape[1]), ...]

    # plt.subplot(121)
    # im = tg.visu.visu2d.scalar2d._scalarfield(
    #     field_scalars=G_selfEp[:, 1, 1].imag, positions=r_probe["r_probe"], cmap="jet"
    # )
    # plt.colorbar(im)
    # plt.subplot(122)
    # im = tg.visu.visu2d.scalar2d._scalarfield(
    #     field_scalars=G_selfHm[:, 1, 1].imag, positions=r_probe["r_probe"], cmap="jet"
    # )
    # plt.colorbar(im)
    # plt.show()

    # # %%
    # # G = G.reshape(len(r_probe), len(r_source), 3,3)
    # # G = torch.moveaxis(G, 2, 3)
    # # print(G)
    # torch.set_printoptions(precision=3, linewidth=150)

    # print(G_Ep.shape)
    # G_self = G_Ep[torch.arange(G_Ep.shape[0]), torch.arange(G_Ep.shape[1]), ...]
    # print(G_self.shape)

    # r_source = torch.as_tensor(r_source)
    # r_probe = torch.as_tensor(r_probe)
    # G_vac = env.get_G_Ep(r_probe, r_source, wavelength)
    # G_vac_6x6 = env.get_G_6x6(r_probe, r_source, wavelength)
    # print(G_vac_6x6.shape)

    # # test if vacuum solution is indentical
    # print(G_6x6[0, 0])
    # print(G_vac_6x6[0])
    # torch.testing.assert_close(G_6x6[0, 0], G_vac_6x6[0], rtol=1e-9, atol=1e-9)



# -*- coding: utf-8 -*-
"""surface discretization for infinitely long 2D geometries

geometry-definitions and discretizer
"""
import warnings

import torch

from torchgdm.constants import DTYPE_FLOAT, DTYPE_COMPLEX
from torchgdm.tools.misc import get_default_device


# ==============================================================================
# geometry definitions
# ==============================================================================
def rectangle(l: int, h: int):
    """rectangle with dimensions `l` along x, `h` along z

    Args:
        l (int): length (units of step)
        h (int): height (units of step)

    Returns:
        func: discretization condition
        tuple of int: discretizer walk limits
    """
    from torchgdm.struct.volume import cuboid

    condition_3d, lim_3d = cuboid(l=l, w=1, h=h)

    def condition(xi, zi):
        return condition_3d(xi, 0, zi)

    mesh_limits = (lim_3d[0], lim_3d[1], lim_3d[4], lim_3d[5])

    return condition, mesh_limits


def square(l: int):
    """square with side length `l` (units of step)"""
    return rectangle(l, l)


def circle(r: float):
    """circle with radius `r` (units of step)"""
    from torchgdm.struct.volume import disc

    condition_3d, lim_3d = disc(r=r, h=1)

    def condition(xi, zi):
        return condition_3d(xi, zi, 0)

    mesh_limits = (lim_3d[0], lim_3d[1], lim_3d[2], lim_3d[3])

    return condition, mesh_limits


def split_ring(r_out: float, r_in: float, alpha_g: int = 0):
    """split ring

    Args:
        r_out (float): outer radius (units of step)
        r_in (float): inner radius (units of step)
        alpha_g (int, optional): split gap angle (rad). Defaults to 0.

    Returns:
        func: discretization condition
        tuple of int: discretizer walk limits
    """

    def condition(xi, zi):
        C1 = r_out**2 >= (xi**2 + zi**2) >= r_in**2
        if zi > 0 and alpha_g != 0:
            C2 = not (
                torch.arctan2(torch.as_tensor(abs(xi)), torch.as_tensor(abs(zi)))
                <= alpha_g / 2.0
            )
        else:
            C2 = True

        return C1 and C2

    mesh_limits = (
        -int(r_out) - 2,
        int(r_out) + 2,
        -int(r_out) - 2,
        int(r_out) + 2,
    )

    return condition, mesh_limits


def triangle_equilateral(d: int, truncate_edge: float = 0.0):
    """Generate uniform right triangular prism

    Args:
        d (int): side length (units of step)
        truncate_edge (float, optional): truncation ratio for the top triangle edge. Defaults to 0.0.

    Returns:
        func: condition(x,z): whether to put or not a mesh-cell at coordinate
        tuple: discretizer walk limits necessary to contain full structure
    """
    from torchgdm.struct.volume import prism_trigonal

    condition_3d, lim_3d = prism_trigonal(d=d, h=1, truncate_edge=truncate_edge)

    def condition(xi, zi):
        return condition_3d(xi, zi, 0)

    mesh_limits = (lim_3d[0], lim_3d[1], lim_3d[2], lim_3d[3])

    return condition, mesh_limits


# ==============================================================================
# discretizer
# ==============================================================================
def discretizer_square(
    condition_func, limits: tuple, z_offset: float = 0.5, step: float = 1.0
):
    """discretize surface of a 2D structure on a square grid

    Args:
        condition_func (_type_): functional, returning the condition as function of coordinate indices. 'condition_func(xi,zi)', returns a bool. (True: put a meshpoint, False: no meshpoint)
        limits (tuple): discretizer walk range. tuple of 6  ints (x0, x1, z0, z1)
        z_offset (float, optional): additional z-offset, to avoid placing dipoles at z=0 (units of step). Defaults to 0.5.
        step (float, optional): nominal stepsize in nm. Defaults to 1.0.

    Returns:
        torch.Tensor: 3d coordinates ([x, 0, z]) of discretization, where y=0. shape (N, 3) for `N` meshpoints.
    """

    sp = []
    for xi in range(int(limits[0]), int(limits[1])):
        for zi in range(int(limits[2]), int(limits[3])):
            if condition_func(xi, zi):
                sp.append([xi, 0, zi + z_offset])

    return torch.as_tensor(sp, dtype=DTYPE_FLOAT) * step


# %%
if __name__ == "__main__":

    l = 10
    w = 7
    h = 3
    step = 10.0
    cond, lim = split_ring(10, 7, torch.pi / 4)
    geometry_pos = discretizer_square(cond, lim, step=step)

    import matplotlib.pyplot as plt

    plt.subplot(aspect="equal")
    plt.scatter(geometry_pos[:, 0], geometry_pos[:, 2])
    plt.show()
    print(len(geometry_pos))
